#
# Copyright (c) 2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import argparse
import fnmatch
import logging
import os
import sys

import libnmstate
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import LLDP
from libnmstate.schema import VLAN


VLAN_TLV_NAME = "ieee-802-1-vlans"
VLAN_ID = "vid"
VLAN_NAME = "name"
BOND_NAME_PREFIX = "bond"


def main():
    logging.basicConfig(
        format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
        level=logging.DEBUG,
    )

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version", action="store_true", help="Display nmstate version"
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        default=False,
        action="store_true",
        dest="dry_run",
        help=(
            "Generate the network state that is going to be applied and print "
            "it out without applying any change."
        ),
    )
    parser.add_argument(
        "only",
        default="*",
        nargs="?",
        metavar=Interface.KEY,
        help="Use only the specified NICs (comma-separated)",
    )

    args = parser.parse_args()
    if args.version:
        print(libnmstate.__version__)
    else:
        bond_vlans(args)


def bond_vlans(args):
    state = _filter_state(libnmstate.show(), args.only)

    if not state[Interface.KEY]:
        sys.stderr.write("ERROR: No such interface\n")
        return os.EX_USAGE

    vlan_to_ifaces = _identify_connected_vlans(state[Interface.KEY])

    desired_state = {Interface.KEY: []}
    for (vlan_id, vlan_name), ifaces_name in vlan_to_ifaces.items():
        if len(ifaces_name) > 1:
            bond_iface = _generate_bond_iface(vlan_id, ifaces_name)
            vlan_iface = _generate_vlan_iface(
                vlan_name, vlan_id, bond_iface[Interface.NAME]
            )
            desired_state[Interface.KEY].extend([bond_iface, vlan_iface])
        else:
            vlan_iface = _generate_vlan_iface(
                vlan_name, vlan_id, ifaces_name[0]
            )
            desired_state[Interface.KEY].append(vlan_iface)

    if args.dry_run:
        desired_state = libnmstate.PrettyState(desired_state)
        sys.stdout.write(desired_state.yaml)
    else:
        libnmstate.apply(desired_state)


def _filter_state(state, allowlist):
    if allowlist != "*":
        patterns = [p for p in allowlist.split(",")]
        # LLDP-autoconf only cares about Interfaces, will not touch routes or
        # route rules.
        state[Interface.KEY] = _filter_interfaces(state, patterns)

    return {Interface.KEY: state[Interface.KEY]}


def _filter_interfaces(state, patterns):
    """
    return the states for all the interfaces from `state` that match at least
    one of the provided patterns.
    """
    show_interfaces = []

    for interface in state[Interface.KEY]:
        for pattern in patterns:
            if fnmatch.fnmatch(interface[Interface.NAME], pattern):
                show_interfaces.append(interface)
                break
    return show_interfaces


def _identify_connected_vlans(ifaces):
    vlan_to_ifaces = {}
    for iface in ifaces:
        if iface.get(LLDP.CONFIG_SUBTREE, {}).get(LLDP.ENABLED):
            neighbors = iface[LLDP.CONFIG_SUBTREE].get(
                LLDP.NEIGHBORS_SUBTREE, []
            )
            for neighbor in neighbors:
                for tlv in neighbor:
                    if tlv.get(VLAN_TLV_NAME):
                        vlan_id = tlv[VLAN_TLV_NAME][0].get(VLAN_ID)
                        vlan_name = tlv[VLAN_TLV_NAME][0].get(VLAN_NAME)
                        identifier = (vlan_id, vlan_name)
                        existing_vlan_info = vlan_to_ifaces.get(identifier, [])
                        existing_vlan_info.append(iface.get(Interface.NAME))
                        vlan_to_ifaces.update({identifier: existing_vlan_info})

    return vlan_to_ifaces


def _generate_bond_iface(vlan_id, ifaces_name):
    bond_state = {
        Interface.NAME: BOND_NAME_PREFIX + str(vlan_id),
        Interface.TYPE: InterfaceType.BOND,
        Interface.STATE: InterfaceState.UP,
        Bond.CONFIG_SUBTREE: {
            Bond.MODE: BondMode.ROUND_ROBIN,
            Bond.PORT: ifaces_name,
        },
    }

    return bond_state


def _generate_vlan_iface(vlan_name, vlan_id, base_iface):
    vlan_state = {
        Interface.NAME: vlan_name,
        Interface.TYPE: InterfaceType.VLAN,
        Interface.STATE: InterfaceState.UP,
        VLAN.CONFIG_SUBTREE: {
            VLAN.BASE_IFACE: base_iface,
            VLAN.ID: vlan_id,
        },
    }

    return vlan_state



#
# Copyright (c) 2020-2022 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

from libnmstate.schema import Route


IPV4_DEFAULT_GATEWAY_DESTINATION = "0.0.0.0/0"
IPV6_DEFAULT_GATEWAY_DESTINATION = "::/0"

IPV4_EMPTY_NEXT_HOP_ADDRESS = "0.0.0.0"
IPV6_EMPTY_NEXT_HOP_ADDRESS = "::"

LOCAL_ROUTE_TABLE = 255


def nispor_route_state_to_nmstate(np_routes):
    ret = []
    for np_route in np_routes:
        if np_route.oif != "lo" and np_route.table != LOCAL_ROUTE_TABLE:
            ret.extend(_nispor_route_to_nmstate(np_route))
    return ret


def nispor_route_state_to_nmstate_static(np_routes):
    ret = []
    for np_route in np_routes:
        if (
            np_route.oif != "lo"
            and np_route.table != LOCAL_ROUTE_TABLE
            and np_route.scope in ["universe", "link"]
            and np_route.protocol in ["static", "boot"]
        ):
            ret.extend(_nispor_route_to_nmstate(np_route))
    return ret


def _nispor_route_to_nmstate(np_rt):
    if np_rt.dst:
        destination = np_rt.dst
    else:
        destination = (
            IPV6_DEFAULT_GATEWAY_DESTINATION
            if np_rt.address_family == "ipv6"
            else IPV4_DEFAULT_GATEWAY_DESTINATION
        )

    if np_rt.via:
        next_hop = np_rt.via
    elif np_rt.gateway:
        next_hop = np_rt.gateway
    else:
        next_hop = (
            IPV6_EMPTY_NEXT_HOP_ADDRESS
            if np_rt.address_family == "ipv6"
            else IPV4_EMPTY_NEXT_HOP_ADDRESS
        )

    nm_route = {
        Route.TABLE_ID: np_rt.table,
        Route.DESTINATION: destination,
        Route.NEXT_HOP_INTERFACE: np_rt.oif if np_rt.oif else "",
        Route.NEXT_HOP_ADDRESS: next_hop,
        Route.METRIC: np_rt.metric if np_rt.metric else 0,
    }
    np_mp_rts = get_multipath_routes(np_rt)
    if np_mp_rts:
        ret = []
        for np_mp_rt in np_mp_rts:
            nm_route_clone = deepcopy(nm_route)
            nm_route_clone[Route.NEXT_HOP_INTERFACE] = np_mp_rt["iface"]
            nm_route_clone[Route.NEXT_HOP_ADDRESS] = np_mp_rt["via"]
            ret.append(nm_route_clone)
        return ret
    else:
        return [nm_route]


# Instead of bumping nispor dependency version, we just use nispor private
# property temporarily
def get_multipath_routes(np_route):
    return np_route._info.get("multipath")



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from .common import NM


def create_vrf_setting(vrf_iface):
    vrf_setting = NM.SettingVrf.new()
    vrf_setting.props.table = vrf_iface.route_table_id
    return vrf_setting


def is_vrf_table_id_changed(vrf_iface, nm_dev):
    return vrf_iface.route_table_id != nm_dev.props.table



# Copyright 2021 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from .clib_wrapper import retrieve_net_state_json


def show(
    *, kernel_only=False, include_status_data=False, include_secrets=False
):
    return json.loads(
        retrieve_net_state_json(
            kernel_only=kernel_only,
            include_status_data=include_status_data,
            include_secrets=include_secrets,
        )
    )


def show_running_config(include_secrets=False):
    return json.loads(
        retrieve_net_state_json(
            include_secrets=include_secrets,
            running_config_only=True,
        )
    )



#
# Copyright (c) 2020-2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateValueError
from libnmstate.schema import VXLAN

from .base_iface import BaseIface


class VxlanIface(BaseIface):
    def __init__(self, info, save_to_disk=True):
        super().__init__(info, save_to_disk)
        self._vxlan_id_changed = False

    @property
    def parent(self):
        return self._vxlan_config.get(VXLAN.BASE_IFACE)

    @property
    def need_parent(self):
        return True

    @property
    def _vxlan_config(self):
        return self.raw.get(VXLAN.CONFIG_SUBTREE, {})

    @property
    def vxlan_id(self):
        return self._vxlan_config.get(VXLAN.ID)

    @property
    def is_vxlan_id_changed(self):
        return self._vxlan_id_changed

    @property
    def is_virtual(self):
        return True

    @property
    def can_have_ip_as_port(self):
        return False

    def _mark_vxlan_id_changed(self, ifaces):
        if self.is_up:
            cur_iface = ifaces.get_cur_iface(self.name, self.type)
            if cur_iface and self.vxlan_id != cur_iface.vxlan_id:
                self._vxlan_id_changed = True

    def gen_metadata(self, ifaces):
        self._mark_vxlan_id_changed(ifaces)
        super().gen_metadata(ifaces)

    def pre_edit_validation_and_cleanup(self):
        if self.is_up:
            self._validate_mandatory_properties()
        super().pre_edit_validation_and_cleanup()

    def _validate_mandatory_properties(self):
        for prop in (VXLAN.ID, VXLAN.BASE_IFACE, VXLAN.REMOTE):
            if prop not in self._vxlan_config:
                raise NmstateValueError(
                    f"Vxlan tunnel {self.name} has missing mandatory "
                    f"property: {prop}"
                )



#
# Copyright (c) 2020-2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateVerificationError
from libnmstate.schema import Ethernet
from libnmstate.schema import InterfaceType

from .base_iface import BaseIface


BNXT_DRIVER_PHYS_PORT_PREFIX = "p"
MULTIPORT_PCI_DEVICE_PREFIX = "n"


class EthernetIface(BaseIface):
    VF_IFACE_NAME_METADATA = "_vf_iface_name"

    def __init__(self, info, save_to_disk=True):
        super().__init__(info, save_to_disk)
        self._is_peer = False

    def merge(self, other):
        super().merge(other)
        EthernetIface._canonicalize(self._info)

    @staticmethod
    def _canonicalize(state):
        """
        * AUTO_NEGOTIATION: true: Remove speed and duplex
        * AUTO_NEGOTIATION: false: copy speed/duplex from other if not defined
        """
        if state.get(Ethernet.CONFIG_SUBTREE, {}).get(
            Ethernet.AUTO_NEGOTIATION
        ):
            state.get(Ethernet.CONFIG_SUBTREE, {}).pop(Ethernet.SPEED, None)
            state.get(Ethernet.CONFIG_SUBTREE, {}).pop(Ethernet.DUPLEX, None)

    def state_for_verify(self):
        state = super().state_for_verify()
        _capitalize_sriov_vf_mac(state)
        EthernetIface._canonicalize(state)
        return state

    @property
    def sriov_total_vfs(self):
        return (
            self.raw.get(Ethernet.CONFIG_SUBTREE, {})
            .get(Ethernet.SRIOV_SUBTREE, {})
            .get(Ethernet.SRIOV.TOTAL_VFS, 0)
        )

    @property
    def sriov_vfs(self):
        return (
            self.raw.get(Ethernet.CONFIG_SUBTREE, {})
            .get(Ethernet.SRIOV_SUBTREE, {})
            .get(Ethernet.SRIOV.VFS_SUBTREE, [])
        )

    @property
    def is_peer(self):
        return self._is_peer

    @property
    def is_sriov(self):
        return self.raw.get(Ethernet.CONFIG_SUBTREE, {}).get(
            Ethernet.SRIOV_SUBTREE
        )

    @property
    def speed(self):
        return self.raw.get(Ethernet.CONFIG_SUBTREE, {}).get(Ethernet.SPEED)

    @property
    def auto_negotiation(self):
        return self.raw.get(Ethernet.CONFIG_SUBTREE, {}).get(
            Ethernet.AUTO_NEGOTIATION
        )

    @property
    def duplex(self):
        return self.raw.get(Ethernet.CONFIG_SUBTREE, {}).get(Ethernet.DUPLEX)

    def remove_vfs_entry_when_total_vfs_decreased(self):
        vfs_count = len(
            self._info[Ethernet.CONFIG_SUBTREE]
            .get(Ethernet.SRIOV_SUBTREE, {})
            .get(Ethernet.SRIOV.VFS_SUBTREE, [])
        )
        if vfs_count > self.sriov_total_vfs:
            [
                self._info[Ethernet.CONFIG_SUBTREE][Ethernet.SRIOV_SUBTREE][
                    Ethernet.SRIOV.VFS_SUBTREE
                ].pop()
                for _ in range(self.sriov_total_vfs, vfs_count)
            ]

    def get_delete_vf_interface_names(self, old_sriov_total_vfs):
        return [
            f"{self.name}v{i}"
            for i in range(self.sriov_total_vfs, old_sriov_total_vfs)
        ]

    def check_total_vfs_matches_vf_list(self, total_vfs):
        return total_vfs == len(self.sriov_vfs)

    def to_dict(self):
        info = super().to_dict()
        for vf_info in (
            info.get(Ethernet.CONFIG_SUBTREE, {})
            .get(Ethernet.SRIOV_SUBTREE, {})
            .get(Ethernet.SRIOV.VFS_SUBTREE, [])
        ):
            vf_info.pop(EthernetIface.VF_IFACE_NAME_METADATA, None)
        return info


def _capitalize_sriov_vf_mac(state):
    vfs = (
        state.get(Ethernet.CONFIG_SUBTREE, {})
        .get(Ethernet.SRIOV_SUBTREE, {})
        .get(Ethernet.SRIOV.VFS_SUBTREE, [])
    )
    for vf in vfs:
        vf_mac = vf.get(Ethernet.SRIOV.VFS.MAC_ADDRESS)
        if vf_mac:
            vf[Ethernet.SRIOV.VFS.MAC_ADDRESS] = vf_mac.upper()


def verify_sriov_vf(iface, cur_ifaces):
    """
    Checking whether VF interface is been created
    """
    if not (
        iface.is_up
        and (iface.is_desired or iface.is_changed)
        and iface.type == InterfaceType.ETHERNET
        and iface.sriov_total_vfs > 0
        and iface.original_desire_dict.get(Ethernet.CONFIG_SUBTREE, {}).get(
            Ethernet.SRIOV_SUBTREE
        )
    ):
        return
    cur_iface = cur_ifaces.get_iface(iface.name, iface.type)
    if not cur_iface:
        # Other verification will raise proper exception when current interface
        # is missing
        return
    cur_vf_names = []
    for sriov_vf in cur_iface.sriov_vfs:
        if sriov_vf.get(EthernetIface.VF_IFACE_NAME_METADATA):
            cur_vf_names.append(sriov_vf[EthernetIface.VF_IFACE_NAME_METADATA])

    if len(cur_vf_names) != iface.sriov_total_vfs:
        raise NmstateVerificationError(
            f"Found VF ports count does not match desired "
            f"{iface.sriov_total_vfs}, current is: {','.join(cur_vf_names)}"
        )
    for cur_vf_name in cur_vf_names:
        if not cur_ifaces.get_iface(cur_vf_name, InterfaceType.ETHERNET):
            raise NmstateVerificationError(
                f"VF interface {cur_vf_name} of PF {iface.name} not found"
            )



import setuptools


def requirements():
    req = []
    with open("requirements.txt") as fd:
        for line in fd:
            line.strip()
            if not line.startswith("#"):
                req.append(line)
    return req


setuptools.setup(
    name="nmstate",
    version="1.3.1",
    author="Gris Ge",
    author_email="fge@redhat.com",
    description="Python binding of nmstate",
    long_description="Python binding of nmstate",
    url="https://github.com/nmstate/nmstate/",
    packages=setuptools.find_packages(),
    install_requires=requirements(),
    license="ASL2.0+",
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
    ],
)



# Copyright 2021-2022 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import c_int, c_char_p, c_uint32, POINTER, byref, cdll
import json
import logging

from .error import (
    NmstateDependencyError,
    NmstateError,
    NmstateInternalError,
    NmstateKernelIntegerRoundedError,
    NmstateNotImplementedError,
    NmstateNotSupportedError,
    NmstatePluginError,
    NmstateValueError,
    NmstateVerificationError,
)

lib = cdll.LoadLibrary("libnmstate.so.1")

lib.nmstate_net_state_retrieve.restype = c_int
lib.nmstate_net_state_retrieve.argtypes = (
    c_uint32,
    POINTER(c_char_p),
    POINTER(c_char_p),
    POINTER(c_char_p),
    POINTER(c_char_p),
)

lib.nmstate_cstring_free.restype = None
lib.nmstate_cstring_free.argtypes = (c_char_p,)

NMSTATE_FLAG_NONE = 0
NMSTATE_FLAG_KERNEL_ONLY = 1 << 1
NMSTATE_FLAG_NO_VERIFY = 1 << 2
NMSTATE_FLAG_INCLUDE_STATUS_DATA = 1 << 3
NMSTATE_FLAG_INCLUDE_SECRETS = 1 << 4
NMSTATE_FLAG_NO_COMMIT = 1 << 5
NMSTATE_FLAG_MEMORY_ONLY = 1 << 6
NMSTATE_FLAG_RUNNING_CONFIG_ONLY = 1 << 7
NMSTATE_PASS = 0


def retrieve_net_state_json(
    kernel_only=False,
    include_status_data=False,
    include_secrets=False,
    running_config_only=False,
):
    c_err_msg = c_char_p()
    c_err_kind = c_char_p()
    c_state = c_char_p()
    c_log = c_char_p()
    flags = NMSTATE_FLAG_NONE
    if kernel_only:
        flags |= NMSTATE_FLAG_KERNEL_ONLY
    if include_status_data:
        flags |= NMSTATE_FLAG_INCLUDE_STATUS_DATA
    if include_secrets:
        flags |= NMSTATE_FLAG_INCLUDE_SECRETS
    if running_config_only:
        flags |= NMSTATE_FLAG_RUNNING_CONFIG_ONLY

    rc = lib.nmstate_net_state_retrieve(
        flags,
        byref(c_state),
        byref(c_log),
        byref(c_err_kind),
        byref(c_err_msg),
    )
    state = c_state.value
    err_msg = c_err_msg.value
    err_kind = c_err_kind.value
    parse_log(c_log.value)
    lib.nmstate_cstring_free(c_log)
    lib.nmstate_cstring_free(c_state)
    lib.nmstate_cstring_free(c_err_kind)
    lib.nmstate_cstring_free(c_err_msg)
    if rc != NMSTATE_PASS:
        raise NmstateError(f"{err_kind}: {err_msg}")
    return state.decode("utf-8")


def apply_net_state(
    state,
    kernel_only=False,
    verify_change=True,
    save_to_disk=True,
    commit=True,
    rollback_timeout=60,
):
    c_err_msg = c_char_p()
    c_err_kind = c_char_p()
    c_state = c_char_p(json.dumps(state).encode("utf-8"))
    c_log = c_char_p()
    flags = NMSTATE_FLAG_NONE
    if kernel_only:
        flags |= NMSTATE_FLAG_KERNEL_ONLY

    if not verify_change:
        flags |= NMSTATE_FLAG_NO_VERIFY

    if not commit:
        flags |= NMSTATE_FLAG_NO_COMMIT

    if not save_to_disk:
        flags |= NMSTATE_FLAG_MEMORY_ONLY

    rc = lib.nmstate_net_state_apply(
        flags,
        c_state,
        rollback_timeout,
        byref(c_log),
        byref(c_err_kind),
        byref(c_err_msg),
    )
    err_msg = c_err_msg.value
    err_kind = c_err_kind.value
    parse_log(c_log.value)
    lib.nmstate_cstring_free(c_log)
    lib.nmstate_cstring_free(c_err_kind)
    lib.nmstate_cstring_free(c_err_msg)
    if rc != NMSTATE_PASS:
        raise map_error(err_kind, err_msg)


def commit_checkpoint(checkpoint):
    c_err_msg = c_char_p()
    c_err_kind = c_char_p()
    c_checkpoint = c_char_p(checkpoint)
    c_log = c_char_p()

    rc = lib.nmstate_checkpoint_commit(
        c_checkpoint,
        byref(c_log),
        byref(c_err_kind),
        byref(c_err_msg),
    )

    err_msg = c_err_msg.value
    err_kind = c_err_kind.value
    parse_log(c_log.value)
    lib.nmstate_cstring_free(c_log)
    lib.nmstate_cstring_free(c_err_kind)
    lib.nmstate_cstring_free(c_err_msg)
    if rc != NMSTATE_PASS:
        raise map_error(err_kind, err_msg)


def rollback_checkpoint(checkpoint):
    c_err_msg = c_char_p()
    c_err_kind = c_char_p()
    c_checkpoint = c_char_p(checkpoint)
    c_log = c_char_p()

    rc = lib.nmstate_checkpoint_rollback(
        c_checkpoint,
        byref(c_log),
        byref(c_err_kind),
        byref(c_err_msg),
    )

    err_msg = c_err_msg.value
    err_kind = c_err_kind.value
    parse_log(c_log.value)
    lib.nmstate_cstring_free(c_log)
    lib.nmstate_cstring_free(c_err_kind)
    lib.nmstate_cstring_free(c_err_msg)
    if rc != NMSTATE_PASS:
        raise map_error(err_kind, err_msg)


def gen_conf(state):
    c_err_msg = c_char_p()
    c_err_kind = c_char_p()
    c_state = c_char_p(json.dumps(state).encode("utf-8"))
    c_configs = c_char_p()
    c_log = c_char_p()
    rc = lib.nmstate_generate_configurations(
        c_state,
        byref(c_configs),
        byref(c_log),
        byref(c_err_kind),
        byref(c_err_msg),
    )
    configs = c_configs.value
    err_msg = c_err_msg.value
    err_kind = c_err_kind.value
    parse_log(c_log.value)
    lib.nmstate_cstring_free(c_log)
    lib.nmstate_cstring_free(c_err_kind)
    lib.nmstate_cstring_free(c_err_msg)
    if rc != NMSTATE_PASS:
        raise map_error(err_kind, err_msg)
    return configs.decode("utf-8")


def map_error(err_kind, err_msg):
    err_msg = err_msg.decode("utf-8")
    err_kind = err_kind.decode("utf-8")
    if err_kind == "VerificationError":
        return NmstateVerificationError(err_msg)
    elif err_kind == "InvalidArgument":
        return NmstateValueError(err_msg)
    elif err_kind == "Bug":
        return NmstateInternalError(err_msg)
    elif err_kind == "PluginFailure":
        return NmstatePluginError(err_msg)
    elif err_kind == "NotImplementedError":
        return NmstateNotImplementedError(err_msg)
    elif err_kind == "KernelIntegerRoundedError":
        return NmstateKernelIntegerRoundedError(err_msg)
    elif err_kind == "NotSupportedError":
        return NmstateNotSupportedError(err_msg)
    elif err_kind == "DependencyError":
        return NmstateDependencyError(err_msg)
    else:
        return NmstateError(f"{err_kind}: {err_msg}")


def parse_log(logs):
    if logs is None:
        return

    log_entries = []
    try:
        log_entries = json.loads(logs.decode("utf-8"))
    except Exception:
        pass
    for log_entry in log_entries:
        msg = f"{log_entry['time']}:{log_entry['file']}: {log_entry['msg']}"
        level = log_entry["level"]

        if level == "ERROR":
            logging.error(msg)
        elif level == "WARN":
            logging.warning(msg)
        elif level == "INFO":
            logging.info(msg)
        else:
            logging.debug(msg)



from .clib_wrapper import NmstateError
from .gen_conf import generate_configurations
from .netapplier import apply
from .netapplier import commit
from .netapplier import rollback
from .netinfo import show
from .netinfo import show_running_config
from .prettystate import PrettyState

__all__ = [
    "NmstateError",
    "PrettyState",
    "apply",
    "commit",
    "generate_configurations",
    "rollback",
    "show",
    "show_running_config",
]

__version__ = "1.3.1"

BASE_ON_RUST = True



#
# Copyright (c) 2020-2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from distutils.version import StrictVersion
import logging
from operator import itemgetter

from libnmstate.error import NmstateDependencyError
from libnmstate.error import NmstateNotSupportedError
from libnmstate.error import NmstateValueError
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import LLDP
from libnmstate.plugin import NmstatePlugin


from .checkpoint import CheckPoint
from .checkpoint import get_checkpoints
from .common import NM
from .context import NmContext
from .device import get_device_common_info
from .device import get_iface_type
from .device import is_kernel_iface
from .device import list_devices
from .dns import get_running as get_dns_running
from .dns import get_running_config as get_dns_running_config
from .infiniband import get_info as get_infiniband_info
from .ipv4 import get_info as get_ipv4_info
from .ipv6 import get_info as get_ipv6_info
from .lldp import get_info as get_lldp_info
from .macvlan import get_current_macvlan_type
from .ovs import get_interface_info as get_ovs_interface_info
from .ovs import get_ovs_bridge_info
from .ovs import get_ovsdb_external_ids
from .ovs import has_ovs_capability
from .profiles import NmProfiles
from .profiles import get_all_applied_configs
from .team import get_info as get_team_info
from .team import has_team_capability
from .translator import Nm2Api
from .user import get_info as get_user_info
from .veth import get_current_veth_type
from .ieee_802_1x import get_802_1x_info


class NetworkManagerPlugin(NmstatePlugin):
    def __init__(self):
        self._ctx = None
        self._checkpoint = None
        self.__kernel_nic_applied_configs = None
        self.__userspace_nic_applied_configs = None

    @property
    def priority(self):
        return NmstatePlugin.DEFAULT_PRIORITY

    @property
    def name(self):
        return "NetworkManager"

    def unload(self):
        if self._ctx:
            self._ctx.clean_up()
            self._ctx = None

    @property
    def _kernel_nic_applied_configs(self):
        if (
            self.__kernel_nic_applied_configs is None
            or self.__userspace_nic_applied_configs is None
        ):
            (
                self.__kernel_nic_applied_configs,
                self.__userspace_nic_applied_configs,
            ) = get_all_applied_configs(self.context)
        return self.__kernel_nic_applied_configs

    @property
    def _userspace_nic_applied_configs(self):
        if (
            self.__kernel_nic_applied_configs is None
            or self.__userspace_nic_applied_configs is None
        ):
            (
                self.__kernel_nic_applied_configs,
                self.__userspace_nic_applied_configs,
            ) = get_all_applied_configs(self.context)
        return self.__userspace_nic_applied_configs

    @property
    def checkpoint(self):
        return self._checkpoint

    @property
    def client(self):
        return self.context.client if self.context else None

    @property
    def context(self):
        if not self._ctx:
            self._ctx = NmContext()
            self._check_version_mismatch()
        return self._ctx

    @property
    def capabilities(self):
        capabilities = []
        if has_ovs_capability(self.client):
            capabilities.append(NmstatePlugin.OVS_CAPABILITY)
        if has_team_capability(self.client):
            capabilities.append(NmstatePlugin.TEAM_CAPABILITY)
        return capabilities

    @property
    def plugin_capabilities(self):
        return [
            NmstatePlugin.PLUGIN_CAPABILITY_IFACE,
            NmstatePlugin.PLUGIN_CAPABILITY_ROUTE,
            NmstatePlugin.PLUGIN_CAPABILITY_ROUTE_RULE,
            NmstatePlugin.PLUGIN_CAPABILITY_DNS,
        ]

    def get_interfaces(self):
        info = []

        devices_info = [
            (dev, get_device_common_info(dev))
            for dev in list_devices(self.client)
        ]

        for dev, devinfo in devices_info:
            if not dev.get_managed():
                # Skip unmanaged interface
                continue
            if is_kernel_iface(dev):
                applied_config = self._kernel_nic_applied_configs.get(
                    dev.get_iface()
                )
            else:
                iface_type = get_iface_type(dev)
                applied_config = self._userspace_nic_applied_configs.get(
                    f"{dev.get_iface()}{iface_type}"
                )

            nm_ac = dev.get_active_connection()
            if (
                nm_ac
                and nm_ac.get_state_flags() & NM.ActivationStateFlags.EXTERNAL
            ):
                # Skip external managed interface
                continue

            iface_info = Nm2Api.get_common_device_info(devinfo)

            act_con = dev.get_active_connection()
            iface_info[Interface.IPV4] = get_ipv4_info(act_con, applied_config)
            iface_info[Interface.IPV6] = get_ipv6_info(act_con, applied_config)
            iface_info.update(get_user_info(self.context, dev))
            iface_info.update(get_lldp_info(self.client, dev))
            iface_info.update(get_team_info(dev))
            iface_info.update(get_infiniband_info(applied_config))
            iface_info.update(get_current_macvlan_type(applied_config))
            iface_info.update(get_current_veth_type(applied_config))
            iface_info.update(get_802_1x_info(self.context, act_con))

            if iface_info[Interface.TYPE] == InterfaceType.OVS_BRIDGE:
                iface_info.update(get_ovs_bridge_info(dev))
                iface_info = _remove_ovs_bridge_unsupported_entries(iface_info)
            elif iface_info[Interface.TYPE] == InterfaceType.OVS_INTERFACE:
                iface_info.update(get_ovs_interface_info(act_con))
            elif iface_info[Interface.TYPE] == InterfaceType.OVS_PORT:
                continue

            if applied_config:
                iface_info.update(get_ovsdb_external_ids(applied_config))

            info.append(iface_info)

        info.sort(key=itemgetter("name"))

        return info

    def get_running_config_interfaces(self):
        iface_infos = self.get_interfaces()
        # Remove LLDP neighber information
        for iface_info in iface_infos:
            if LLDP.CONFIG_SUBTREE in iface_info:
                iface_info[LLDP.CONFIG_SUBTREE].pop(
                    LLDP.NEIGHBORS_SUBTREE, None
                )
        return iface_infos

    def get_routes(self):
        return {}

    def get_route_rules(self):
        """
        Nispor will provide running config of route rule from kernel.
        """
        return {}

    def get_dns_client_config(self):
        return {
            DNS.RUNNING: get_dns_running(self.client),
            DNS.CONFIG: get_dns_running_config(
                self._kernel_nic_applied_configs
            ),
        }

    def refresh_content(self):
        self.__kernel_nic_applied_configs = None
        self.__userspace_nic_applied_configs = None
        if self._ctx:
            self._ctx.refresh()

    def apply_changes(self, net_state, save_to_disk):
        NmProfiles(self.context).apply_config(net_state, save_to_disk)

    def _load_checkpoint(self, checkpoint_path):
        if checkpoint_path:
            if self._checkpoint:
                # Old checkpoint might timeout, hence it's legal to load
                # another one.
                self._checkpoint.clean_up()
            candidates = get_checkpoints(self.client)
            if checkpoint_path in candidates:
                self._checkpoint = CheckPoint(
                    nm_context=self.context, dbuspath=checkpoint_path
                )
            else:
                raise NmstateValueError("No checkpoint specified or found")
        else:
            if not self._checkpoint:
                # Get latest one
                candidates = get_checkpoints(self.client)
                if candidates:
                    self._checkpoint = CheckPoint(
                        nm_context=self.context, dbuspath=candidates[0]
                    )
                else:
                    raise NmstateValueError("No checkpoint specified or found")

    def create_checkpoint(self, timeout=60):
        self._checkpoint = CheckPoint.create(self.context, timeout)
        return str(self._checkpoint)

    def rollback_checkpoint(self, checkpoint=None):
        self._load_checkpoint(checkpoint)
        self._checkpoint.rollback()
        self._checkpoint = None

    def destroy_checkpoint(self, checkpoint=None):
        self._load_checkpoint(checkpoint)
        self._checkpoint.destroy()
        self._checkpoint = None

    def _check_version_mismatch(self):
        nm_client_version = self.client.get_version()
        nm_utils_version = _nm_utils_decode_version()

        if nm_client_version is None:
            raise NmstateDependencyError(
                "NetworkManager daemon is not running which is required for "
                "NetworkManager plugin"
            )
        elif StrictVersion(nm_client_version) != StrictVersion(
            nm_utils_version
        ):
            logging.warning(
                "libnm version %s mismatches NetworkManager version %s",
                nm_utils_version,
                nm_client_version,
            )

        logging.debug(f"NetworkManager version {nm_client_version}")

    def generate_configurations(self, net_state):
        if not hasattr(NM, "keyfile_write"):
            raise NmstateNotSupportedError(
                f"Current NetworkManager version does not support generating "
                "configurations, please upgrade to 1.30 or later versoin."
            )
        return NmProfiles(None).generate_config_strings(net_state)

    def get_ignored_kernel_interface_names(self):
        """
        Return a list of unmanged kernel interface names.
        """
        ignored_ifaces = set()
        for nm_dev in list_devices(self.client):
            if (
                nm_dev
                and nm_dev.get_iface()
                and not nm_dev.get_managed()
                and is_kernel_iface(nm_dev)
            ):
                ignored_ifaces.add(nm_dev.get_iface())
        return list(ignored_ifaces)


def _remove_ovs_bridge_unsupported_entries(iface_info):
    """
    OVS bridges are not supporting several common interface key entries.
    These entries are removed explicitly.
    """
    iface_info.pop(Interface.IPV4, None)
    iface_info.pop(Interface.IPV6, None)
    iface_info.pop(Interface.MTU, None)

    return iface_info


def _nm_utils_decode_version():
    return f"{NM.MAJOR_VERSION}.{NM.MINOR_VERSION}.{NM.MICRO_VERSION}"



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.ifaces import NmstateLinuxBridgePortVlan

from .common import NM


def nmstate_port_vlan_to_nm(nmstate_vlan_config):
    nm_vlans = []
    nmstate_port_vlan = NmstateLinuxBridgePortVlan(nmstate_vlan_config)
    for kernel_vlan in nmstate_port_vlan.to_kernel_vlans():
        nm_vlan = NM.BridgeVlan.new(kernel_vlan.vid_min, kernel_vlan.vid_max)
        nm_vlan.set_untagged(kernel_vlan.is_egress_untagged)
        nm_vlan.set_pvid(kernel_vlan.is_pvid)
        nm_vlans.append(nm_vlan)
    return nm_vlans



# Copyright 2021 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .clib_wrapper import apply_net_state
from .clib_wrapper import commit_checkpoint
from .clib_wrapper import rollback_checkpoint


def apply(
    desired_state,
    *,
    kernel_only=False,
    verify_change=True,
    save_to_disk=True,
    commit=True,
    rollback_timeout=60,
):
    return apply_net_state(
        desired_state,
        kernel_only=kernel_only,
        verify_change=verify_change,
        save_to_disk=save_to_disk,
        commit=commit,
        rollback_timeout=rollback_timeout,
    )


def commit(*, checkpoint=None):
    commit_checkpoint(checkpoint)


def rollback(*, checkpoint=None):
    rollback_checkpoint(checkpoint)



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import time
import logging

from libnmstate.error import NmstateLibnmError
from libnmstate.error import NmstateDependencyError
from libnmstate.error import NmstateInternalError

from .common import GLib
from .common import Gio
from .common import NM
from .device import get_nm_dev
from .device import get_iface_type
from .ipv4 import is_dynamic as is_ipv4_dynamic
from .ipv6 import is_dynamic as is_ipv6_dynamic


NM_AC_STATE_CHANGED_SIGNAL = "state-changed"
FALLBACK_CHECKER_INTERNAL = 15
MAX_OVS_IFACE_PREPARE_TIME = FALLBACK_CHECKER_INTERNAL * 2
GIO_ERROR_DOMAIN = "g-io-error-quark"
ACTIVATION_RETRY_SLEEP = 5


def is_activated(nm_ac, nm_dev):
    if not (nm_ac and nm_dev):
        return False

    state = nm_ac.get_state()
    if state == NM.ActiveConnectionState.ACTIVATED:
        return True
    elif state == NM.ActiveConnectionState.ACTIVATING:
        # OVS bridge and OVS port are not allowed to have IP, hence we wait it
        # to reach ACTIVATED state.
        if nm_dev.get_device_type() in (
            NM.DeviceType.OVS_PORT,
            NM.DeviceType.OVS_BRIDGE,
        ):
            return False
        ac_state_flags = nm_ac.get_state_flags()
        nm_flags = NM.ActivationStateFlags
        ip4_is_dynamic = is_ipv4_dynamic(nm_ac)
        ip6_is_dynamic = is_ipv6_dynamic(nm_ac)
        if (
            ac_state_flags & nm_flags.IS_MASTER
            or (ip4_is_dynamic and ac_state_flags & nm_flags.IP6_READY)
            or (ip6_is_dynamic and ac_state_flags & nm_flags.IP4_READY)
            or (ip4_is_dynamic and ip6_is_dynamic)
        ):
            # For interface meet any condition below will be
            # treated as activated when reach IP_CONFIG state:
            #   * Is controller device.
            #   * DHCPv4 enabled with IP6_READY flag.
            #   * DHCPv6/Autoconf with IP4_READY flag.
            #   * DHCPv4 enabled with DHCPv6/Autoconf enabled.
            return (
                NM.DeviceState.IP_CONFIG
                <= nm_dev.get_state()
                <= NM.DeviceState.ACTIVATED
            )

    return False


def is_activating(nm_ac, nm_dev):
    if not nm_ac or not nm_dev:
        return True
    if nm_dev.get_state_reason() == NM.DeviceStateReason.NEW_ACTIVATION:
        return True

    return (
        nm_ac.get_state() == NM.ActiveConnectionState.ACTIVATING
    ) and not is_activated(nm_ac, nm_dev)


class ProfileActivation:
    def __init__(self, ctx, iface_name, iface_type, nm_profile, nm_dev):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_ac = None
        self._nm_dev = nm_dev
        self._nm_profile = nm_profile
        self._ac_handlers = set()
        self._dev_handlers = set()
        self._action = None
        self._fallback_checker = None
        self._fallback_checker_counter = 0

    def run(self):
        specific_object = None
        self._action = (
            f"Activate profile uuid:{self._nm_profile.get_uuid()} "
            f"iface:{self._iface_name} type: {self._iface_type}"
        )

        retry = True
        self._ctx.register_async(self._action)
        self._ctx.client.activate_connection_async(
            self._nm_profile,
            self._nm_dev,
            specific_object,
            self._ctx.cancellable,
            self._activate_profile_callback,
            retry,
        )
        self._fallback_checker = GLib.timeout_source_new(
            FALLBACK_CHECKER_INTERNAL * 1000
        )
        self._fallback_checker.set_callback(
            self._fallback_checker_callback, None
        )
        self._fallback_checker.attach(self._ctx.context)

    @staticmethod
    def wait(ctx, nm_ac, nm_dev):
        activation = ProfileActivation(
            ctx,
            nm_dev.get_iface(),
            get_iface_type(nm_dev),
            None,
            nm_dev,
        )
        activation._nm_ac = nm_ac
        activation._action = (
            f"Waiting activation of {activation._iface_name} "
            f"{activation._iface_type}"
        )
        ctx.register_async(activation._action)
        activation._fallback_checker = GLib.timeout_source_new(
            FALLBACK_CHECKER_INTERNAL * 1000
        )
        activation._fallback_checker.set_callback(
            activation._fallback_checker_callback, None
        )
        activation._fallback_checker.attach(ctx.context)
        activation._wait_profile_activation()

    def _activate_profile_callback(self, nm_client, result, retry):
        nm_ac = None
        if self._ctx.is_cancelled():
            self._activation_clean_up()
            return
        try:
            nm_ac = nm_client.activate_connection_finish(result)
        except GLib.Error as e:
            if retry:
                retry = False
                specific_object = None
                logging.debug(
                    f"Action {self._action} failed, trying again in "
                    f"{ACTIVATION_RETRY_SLEEP} seconds."
                )
                time.sleep(ACTIVATION_RETRY_SLEEP)
                self._ctx.client.activate_connection_async(
                    self._nm_profile,
                    self._nm_dev,
                    specific_object,
                    self._ctx.cancellable,
                    self._activate_profile_callback,
                    retry,
                )
                return
            elif e.matches(Gio.io_error_quark(), Gio.IOErrorEnum.TIMED_OUT):
                logging.debug(
                    f"{self._action} timeout on activation, "
                    "using fallback method to wait activation"
                )
                return
            else:
                self._ctx.fail(
                    NmstateLibnmError(f"{self._action} failed: error={e}")
                )
                return
        except Exception as e:
            self._ctx.fail(
                NmstateLibnmError(f"{self._action} failed: error={e}")
            )
            return

        if nm_ac is None:
            self._ctx.fail(
                NmstateLibnmError(
                    f"{self._action} failed: "
                    "error='None return from activate_connection_finish()'"
                )
            )
        else:
            logging.debug(
                f"Connection activation initiated: iface={self._iface_name} "
                f"type={self._iface_type} con-state={nm_ac.get_state()}"
            )
            self._nm_ac = nm_ac
            self._nm_dev = get_nm_dev(
                self._ctx, self._iface_name, self._iface_type
            )
            self._wait_profile_activation()

    def _wait_profile_activation(self):
        if is_activated(self._nm_ac, self._nm_dev):
            logging.debug(
                "Connection activation succeeded: "
                f"iface={self._iface_name}, type={self._iface_type}, "
                f"con_state={self._nm_ac.get_state()}, "
                f"dev_state={self._nm_dev.get_state()}, "
                f"state_flags={self._nm_ac.get_state_flags()}"
            )
            self._activation_clean_up()
            self._ctx.finish_async(self._action)
        elif is_activating(self._nm_ac, self._nm_dev):
            if self._nm_ac:
                self._wait_nm_ac_activation()
            if self._nm_dev:
                self._wait_nm_dev_activation()
            if not self._nm_ac and not self._nm_dev:
                self._ctx.fail(
                    NmstateInternalError(
                        f"{self._action} failed: no nm_ac or nm_dev"
                    )
                )
        else:
            if self._nm_dev:
                error_msg = (
                    f"Connection {self._nm_profile.get_uuid()} failed: "
                    f"state={self._nm_ac.get_state()} "
                    f"reason={self._nm_ac.get_state_reason()} "
                    f"dev_state={self._nm_dev.get_state()} "
                    f"dev_reason={self._nm_dev.get_state_reason()}"
                )
            else:
                error_msg = (
                    f"Connection {self._nm_profile.get_uuid()} failed: "
                    f"state={self._nm_ac.get_state()} "
                    f"reason={self._nm_ac.get_state_reason()} dev=None"
                )
            self._activation_clean_up()
            logging.error(error_msg)
            self._ctx.fail(
                NmstateLibnmError(f"{self._action} failed: {error_msg}")
            )

    def _activation_clean_up(self):
        self._remove_ac_handlers()
        self._remove_dev_handlers()
        if self._fallback_checker:
            self._fallback_checker.destroy()
            self._fallback_checker = None

    def _remove_ac_handlers(self):
        for handler_id in self._ac_handlers:
            self._nm_ac.handler_disconnect(handler_id)
        self._ac_handlers = set()

    def _remove_dev_handlers(self):
        for handler_id in self._dev_handlers:
            self._nm_dev.handler_disconnect(handler_id)
        self._dev_handlers = set()

    def _wait_nm_ac_activation(self):
        user_data = None
        self._ac_handlers.add(
            self._nm_ac.connect(
                NM_AC_STATE_CHANGED_SIGNAL,
                self._ac_state_change_callback,
                user_data,
            )
        )
        self._ac_handlers.add(
            self._nm_ac.connect(
                "notify::state-flags",
                self._ac_state_flags_change_callback,
                user_data,
            )
        )

    def _ac_state_change_callback(self, _nm_ac, _state, _reason, _user_data):
        if self._ctx.is_cancelled():
            self._activation_clean_up()
            return
        self._activation_progress_check()

    def _ac_state_flags_change_callback(self, _nm_ac, _state, _user_data):
        if self._ctx.is_cancelled():
            self._activation_clean_up()
            return
        self._activation_progress_check()

    def _wait_nm_dev_activation(self):
        user_data = None
        self._dev_handlers.add(
            self._nm_dev.connect(
                "state-changed", self._dev_state_change_callback, user_data
            )
        )

    def _dev_state_change_callback(
        self, _nm_dev, _new_state, _old_state, _reason, _user_data
    ):
        if self._ctx.is_cancelled():
            self._activation_clean_up()
            return
        self._activation_progress_check()

    def _activation_progress_check(self):
        cur_nm_dev = get_nm_dev(self._ctx, self._iface_name, self._iface_type)
        if cur_nm_dev and cur_nm_dev != self._nm_dev:
            logging.debug(
                f"The NM.Device of profile {self._iface_name} "
                f"{self._iface_type} changed"
            )
            self._remove_dev_handlers()
            self._nm_dev = cur_nm_dev
            self._wait_nm_dev_activation()

        if cur_nm_dev:
            cur_nm_ac = cur_nm_dev.get_active_connection()
            if cur_nm_ac and cur_nm_ac != self._nm_ac:
                logging.debug(
                    f"Active connection of device {self._iface_name} "
                    "has been replaced"
                )
                self._remove_ac_handlers()
                self._nm_ac = cur_nm_ac
                self._wait_nm_ac_activation()

        if is_activated(self._nm_ac, self._nm_dev):
            logging.debug(
                "Connection activation succeeded: "
                f"iface={self._iface_name}, type={self._iface_type}, "
                f"con_state={self._nm_ac.get_state()}, "
                f"dev_state={self._nm_dev.get_state()}, "
                f"state_flags={self._nm_ac.get_state_flags()}"
            )
            self._activation_clean_up()
            self._ctx.finish_async(self._action)
        elif not is_activating(self._nm_ac, self._nm_dev):
            nm_ac_reason = f"{self._nm_ac.get_state_reason()}"
            nm_dev_reason = None
            if self._nm_dev:
                nm_dev_reason = self._nm_dev.get_state_reason()

            if nm_dev_reason == NM.DeviceStateReason.OVSDB_FAILED:
                error = NmstateDependencyError(
                    f"{self._action} failed: failed to communicating with "
                    f"Open vSwitch database, {nm_dev_reason}"
                )
            else:
                reason = nm_ac_reason + (
                    str(nm_dev_reason) if nm_dev_reason else ""
                )
                error = NmstateLibnmError(
                    f"{self._action} failed: reason={reason}"
                )
            self._activation_clean_up()
            self._ctx.fail(error)

    def _fallback_checker_callback(self, _user_data):
        self._fallback_checker_counter += 1
        nm_dev = get_nm_dev(self._ctx, self._iface_name, self._iface_type)
        if nm_dev:
            self._nm_dev = nm_dev
            self._activation_progress_check()
            # When OVSDB connection is invalid(such as been mounted as
            # /dev/null), NM will hang on the activation of ovs internal
            # interface with state ACITVATING with reason UNKNOWN forever with
            # no state change signal. The fallback check only found it
            # as activating which lead us hang till killed by idle timeout.
            # To prevent that, when we found OVS interface interface in
            # `NM.DeviceState.PREPARE` on in second call of fallbacker,
            # we fail the action as NmstateDependencyError.
            if (
                self._fallback_checker_counter
                >= MAX_OVS_IFACE_PREPARE_TIME / FALLBACK_CHECKER_INTERNAL
                and nm_dev.get_device_type() == NM.DeviceType.OVS_INTERFACE
                and nm_dev.get_state() == NM.DeviceState.PREPARE
            ):
                self._ctx.fail(
                    NmstateDependencyError(
                        f"{self._action} failed: timeout on creating OVS "
                        "interface, please check Open vSwitch daemon"
                    )
                )

        return GLib.SOURCE_CONTINUE


class ActiveConnectionDeactivate:
    def __init__(self, ctx, iface_name, iface_type, nm_ac):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_ac = nm_ac
        self._fallback_checker = None
        self._signal_handler = None

    def run(self):
        if self._nm_ac.props.state == NM.ActiveConnectionState.DEACTIVATED:
            return

        action = f"Deactivate profile: {self._iface_name} {self._iface_type}"
        self._ctx.register_async(action)
        self._signal_handler = self._nm_ac.connect(
            NM_AC_STATE_CHANGED_SIGNAL,
            self._wait_state_changed_callback,
            action,
        )
        if self._nm_ac.props.state != NM.ActiveConnectionState.DEACTIVATING:
            user_data = action
            self._ctx.client.deactivate_connection_async(
                self._nm_ac,
                self._ctx.cancellable,
                self._deactivate_connection_callback,
                user_data,
            )
        self._fallback_checker = GLib.timeout_source_new(
            FALLBACK_CHECKER_INTERNAL * 1000
        )
        self._fallback_checker.set_callback(
            self._fallback_checker_callback, action
        )
        self._fallback_checker.attach(self._ctx.context)

    def _clean_up(self):
        if self._signal_handler:
            if self._nm_ac:
                self._nm_ac.handler_disconnect(self._signal_handler)
            self._signal_handler = None
        if self._fallback_checker:
            self._fallback_checker.destroy()
            self._fallback_checker = None

    def _wait_state_changed_callback(self, nm_ac, state, reason, action):
        if self._ctx.is_cancelled():
            return
        if nm_ac.props.state == NM.ActiveConnectionState.DEACTIVATED:
            logging.debug(
                "Connection deactivation succeeded on %s",
                self._iface_name,
            )
            self._clean_up()
            self._ctx.finish_async(action)

    def _deactivate_connection_callback(self, nm_client, result, user_data):
        action = user_data
        if self._ctx.is_cancelled():
            self._clean_up()
            return

        try:
            success = nm_client.deactivate_connection_finish(result)
        except GLib.Error as e:
            if e.matches(
                NM.ManagerError.quark(), NM.ManagerError.CONNECTIONNOTACTIVE
            ):
                success = True
                logging.debug(
                    "Connection is not active on {}, no need to "
                    "deactivate".format(self._iface_name)
                )
                self._clean_up()
                self._ctx.finish_async(action)
            elif e.matches(Gio.io_error_quark(), Gio.IOErrorEnum.TIMED_OUT):
                logging.debug(
                    f"{action} timeout, using fallback method to "
                    "wait profile deactivation"
                )
                return
            else:
                self._clean_up()
                self._ctx.fail(
                    NmstateLibnmError(f"{action} failed: error={e}")
                )
                return
        except Exception as e:
            self._clean_up()
            self._ctx.fail(
                NmstateLibnmError(
                    "BUG: Unexpected error when activating "
                    f"{self._iface_name} error={e}"
                )
            )
            return

        if not success:
            self._clean_up()
            self._ctx.fail(
                NmstateLibnmError(
                    f"{action} failed: error='None returned from "
                    "deactivate_connection_finish()'"
                )
            )

    def _fallback_checker_callback(self, action):
        nm_dev = get_nm_dev(self._ctx, self._iface_name, self._iface_type)
        if nm_dev:
            nm_ac = nm_dev.get_active_connection()
            if (
                nm_ac
                and nm_ac.props.state != NM.ActiveConnectionState.DEACTIVATED
            ):
                return GLib.SOURCE_CONTINUE

        self._clean_up()
        self._ctx.finish_async(action)
        return GLib.SOURCE_REMOVE



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateValueError
from libnmstate.schema import MacVlan

from .base_iface import BaseIface


class MacVlanIface(BaseIface):
    @property
    def parent(self):
        return self.config_subtree.get(MacVlan.BASE_IFACE)

    @property
    def need_parent(self):
        return True

    @property
    def config_subtree(self):
        return self.raw.get(MacVlan.CONFIG_SUBTREE, {})

    @property
    def is_virtual(self):
        return True

    @property
    def can_have_ip_when_enslaved(self):
        return False

    @property
    def mode(self):
        return self.config_subtree[MacVlan.MODE]

    @property
    def base_iface(self):
        return self.config_subtree[MacVlan.BASE_IFACE]

    @property
    def promiscuous(self):
        return self.config_subtree.get(MacVlan.PROMISCUOUS)

    def pre_edit_validation_and_cleanup(self):
        if self.is_up:
            self._validate_mode()
            self._validate_mandatory_properties()
        super().pre_edit_validation_and_cleanup()

    def _validate_mode(self):
        if self.config_subtree.get(
            MacVlan.MODE
        ) != MacVlan.Mode.PASSTHRU and not self.config_subtree.get(
            MacVlan.PROMISCUOUS
        ):
            raise NmstateValueError(
                "Disable promiscuous is only allowed on passthru mode"
            )
        if self.config_subtree.get(MacVlan.MODE) == MacVlan.Mode.UNKNOWN:
            raise NmstateValueError(
                "Mode unknown is not supported when appying the state"
            )

    def _validate_mandatory_properties(self):
        for prop in (MacVlan.MODE, MacVlan.BASE_IFACE):
            if prop not in self.config_subtree:
                raise NmstateValueError(
                    f"{self.type} tunnel {self.name} has missing mandatory"
                    f" property: {prop}"
                )
